package filegateway

import (
	"crypto/tls"
	"crypto/x509"
	"encoding/pem"
	"net"
	"os"
	"strings"
	"syscall"
	"testing"

	"github.com/secsy/goftp"
)

var ftpServerCert = `-----BEGIN CERTIFICATE-----
MIICkjCCAfugAwIBAgIUNFDB4EEpjwG0EO+Z3fekFK0XGUwwDQYJKoZIhvcNAQEL
BQAwWjELMAkGA1UEBhMCRlIxDzANBgNVBAgMBkZyYW5jZTETMBEGA1UECgwKWGJ1
cyBUZXN0czElMCMGA1UEAwwceGJ1cy1maWxlZ2F0ZXdheS1mdHBzLXNlcnZlcjAg
Fw0xOTA3MTYxMDAwMTFaGA8yMjkzMDQyOTEwMDAxMVowWjELMAkGA1UEBhMCRlIx
DzANBgNVBAgMBkZyYW5jZTETMBEGA1UECgwKWGJ1cyBUZXN0czElMCMGA1UEAwwc
eGJ1cy1maWxlZ2F0ZXdheS1mdHBzLXNlcnZlcjCBnzANBgkqhkiG9w0BAQEFAAOB
jQAwgYkCgYEAvLMND9DWDkfEThHJvowQ648EaAAB0F0w7QRPrb0hKwCChSrC4viV
cl5PbK5RthO0moYrrjFBUAEBbZrIMXYuaR+9+Im+pSbqe1hq5fs5+0ronyMwstEv
DwwyPyK0ln845ejeQ1VTsanl8xPOQ/SMRh0SSErcG6oT+H3/nY1YIo8CAwEAAaNT
MFEwHQYDVR0OBBYEFAdF8RFOAecLY6LSHUih7YsaTPS4MB8GA1UdIwQYMBaAFAdF
8RFOAecLY6LSHUih7YsaTPS4MA8GA1UdEwEB/wQFMAMBAf8wDQYJKoZIhvcNAQEL
BQADgYEALXco3kxg2aX8sIqqmpSoci3W4vLV0+JBnreI8xesIF4qfgew2SMv2lUy
RjSNWlSFc0kFfHzmYNpSrpTST8qLPSHRfXsWPfo8Sm6VqIQ0T3qCoO40CVNWB5MH
duiFZ0XPojdP6wONAdG91hlVGaQ/7Ph/v5PaNCbDLMf8CZS4eZ4=
-----END CERTIFICATE-----
`

func loadFtpTestServerCert(t *testing.T) *x509.Certificate {
	pemBlock := []byte(ftpServerCert)

	var derBlock *pem.Block
	for {
		derBlock, pemBlock = pem.Decode(pemBlock)
		if derBlock == nil {
			break
		}
		if derBlock.Type == "CERTIFICATE" {
			break
		}
		derBlock = nil
	}
	if derBlock == nil {
		t.Fatal("Found no certificate")
	}
	x509Cert, err := x509.ParseCertificate(derBlock.Bytes)
	if err != nil {
		t.Fatal(err)
	}
	return x509Cert
}

func TestFtpFilesystem(t *testing.T) {
	t.Run("Proftpd", proftpdTetss)
	t.Run("Vsftpd", vsftpdTests)
}

func tryTCPConnect(t *testing.T, hostport string) error {
	c, err := net.Dial("tcp", hostport)
	if err == nil {
		err = c.Close()
	}
	if ne, ok := err.(*net.OpError); ok {
		if se, ok := ne.Err.(*os.SyscallError); ok {
			if ne.Op == "dial" && se.Err == syscall.ECONNREFUSED {
				return err
			}
		}
	}
	if err != nil {
		t.Fatal(err)
	}
	return nil
}
func tryFTPConnect(t *testing.T, hostport string, cfg goftp.Config) error {
	client, err := goftp.DialConfig(cfg, hostport)
	if err != nil {
		t.Fatal(err)
	}
	defer client.Close()

	conn, err := client.OpenRawConn()
	if err != nil {
		if !strings.Contains(err.Error(), "EOF") {
			t.Fatal(err)
		}
		return nil
	}
	return conn.Close()
}

func proftpdTetss(t *testing.T) {
	var host, port string
	certpool := x509.NewCertPool()
	certpool.AddCert(loadFtpTestServerCert(t))
	if IsInCI() {
		host = os.Getenv("PROFTPD_PORT_21_TCP_ADDR")
		port = os.Getenv("PROFTPD_PORT_21_TCP_PORT")
	} else {
		var err error
		port, err = dockerComposePort("proftpd", "21")
		if err != nil {
			t.Fatal(err)
		}
		host = "localhost"
	}

	if err := tryFTPConnect(t, host+":"+port, goftp.Config{
		User:     "ftpuser",
		Password: "ahqueibae5phiraiCo0E",
		TLSConfig: &tls.Config{
			ServerName: "xbus-filegateway-ftps-server",
			RootCAs:    certpool,
		},
		TLSMode: goftp.TLSExplicit,
	}); err != nil {
		t.Fatal(err)
	}

	t.Run("NoTLS", func(t *testing.T) {
		FilesystemTestSuite(t, func(options FilesystemOptions) (Filesystem, error) {
			fs, err := ftpFilesystemFactory(options)
			if err != nil {
				return nil, err
			}
			return fs, nil
		}, FilesystemOptions{
			Scheme:   "ftp",
			Host:     host,
			Port:     port,
			Username: "ftpuser",
			Password: "ahqueibae5phiraiCo0E",
		})
	})

	t.Run("TLS", func(t *testing.T) {
		FilesystemTestSuite(t, func(options FilesystemOptions) (Filesystem, error) {
			fs, err := ftpFilesystemFactory(options)
			if err != nil {
				return nil, err
			}
			return fs, nil
		}, FilesystemOptions{
			Scheme:   "ftps",
			Host:     host,
			Port:     port,
			Username: "ftpuser",
			Password: "ahqueibae5phiraiCo0E",
			Custom: CustomOptions{
				"host-cert":      ftpServerCert,
				"tls-mode":       "explicit",
				"tls-servername": "xbus-filegateway-ftps-server",
			},
		})
	})
}

func vsftpdTests(t *testing.T) {
	var host, port string
	certpool := x509.NewCertPool()
	certpool.AddCert(loadFtpTestServerCert(t))
	if IsInCI() {
		host = os.Getenv("VSFTPD_PORT_21_TCP_ADDR")
		port = os.Getenv("VSFTPD_PORT_21_TCP_PORT")
	} else {
		var err error
		port, err = dockerComposePort("vsftpd", "21")
		if err != nil {
			t.Fatal(err)
		}
		host = "localhost"
	}

	if err := tryFTPConnect(t, host+":"+port, goftp.Config{
		User:     "ftpuser",
		Password: "ahqueibae5phiraiCo0E",
		TLSConfig: &tls.Config{
			ServerName: "xbus-filegateway-ftps-server",
			RootCAs:    certpool,
		},
		TLSMode: goftp.TLSExplicit,
	}); err != nil {
		t.Fatal(err)
	}

	t.Run("NoTLS", func(t *testing.T) {
		FilesystemTestSuite(t, func(options FilesystemOptions) (Filesystem, error) {
			fs, err := ftpFilesystemFactory(options)
			if err != nil {
				return nil, err
			}
			return fs, nil
		}, FilesystemOptions{
			Scheme:   "ftp",
			Host:     host,
			Port:     port,
			Username: "ftpuser",
			Password: "ahqueibae5phiraiCo0E",
			Custom: CustomOptions{
				"flavor": "vsftpd",
			},
		})
	})

	t.Run("TLS", func(t *testing.T) {
		FilesystemTestSuite(t, func(options FilesystemOptions) (Filesystem, error) {
			fs, err := ftpFilesystemFactory(options)
			if err != nil {
				return nil, err
			}
			return fs, nil
		}, FilesystemOptions{
			Scheme:   "ftps",
			Host:     host,
			Port:     port,
			Username: "ftpuser",
			Password: "ahqueibae5phiraiCo0E",
			Custom: CustomOptions{
				"flavor":         "vsftpd",
				"host-cert":      ftpServerCert,
				"tls-mode":       "explicit",
				"tls-servername": "xbus-filegateway-ftps-server",
			},
		})
	})
}
