package stdlib

import (
	"encoding/json"
	"fmt"
	"io"

	"xbus.io/go-xbus/v3/api"
)

// FileMeta ...
type FileMeta struct {
	Filename string       `json:"filename"`
	Type     string       `json:"type,omitempty"`
	Size     int          `json:"size,omitempty"`
	Encoding string       `json:"encoding,omitempty"`
	Location FileLocation `json:"location,omitempty"`
}

// FileLocation holds information on where the actual file data is stored
type FileLocation struct {
	Type string
	Args map[string]interface{}
}

// FileLocationErr wraps errors returned by a FileLocation so we always know
// which location type it is related to
type FileLocationErr struct {
	Type string
	Err  error
}

func (err FileLocationErr) Unwrap() error {
	return err.Err
}

func (err FileLocationErr) Error() string {
	return fmt.Sprintf("file-location %s: %s", err.Type, err.Err)
}

// WrapFileLocationReadCloser returns a wrapped io.ReadCloser so the returned errors are
// all FileLocationErr, except for io.EOF.
func WrapFileLocationReadCloser(flType string, readCloser io.ReadCloser) FileLocationReadCloserWrapper {
	return FileLocationReadCloserWrapper{flType, readCloser}
}

// FileLocationReadCloserWrapper wraps a io.ReadCloser
type FileLocationReadCloserWrapper struct {
	Type       string
	readCloser io.ReadCloser
}

// Read the underlying io.ReadCloser and wraps its errors
func (w FileLocationReadCloserWrapper) Read(b []byte) (int, error) {
	c, err := w.readCloser.Read(b)
	if err != nil && err != io.EOF {
		return c, FileLocationErr{w.Type, err}
	}
	return c, err
}

// Close the underlying io.ReadCloser and wraps its errors
func (w FileLocationReadCloserWrapper) Close() error {
	if err := w.readCloser.Close(); err != nil {
		return FileLocationErr{w.Type, err}
	}
	return nil
}

// StringArg returns a Args as a string
func (fl FileLocation) StringArg(key string) (string, error) {
	if value, ok := fl.Args[key]; ok {
		if s, ok := value.(string); ok {
			return s, nil
		}
		return "", fmt.Errorf("location %s: %s must be a string, got: %s", fl.Type, key, value)
	}
	return "", fmt.Errorf("location %s: missing %s", fl.Type, key)
}

// StringArgD returns a Args as a string
func (fl FileLocation) StringArgD(key string, defaultValue string) (string, error) {
	if value, ok := fl.Args[key]; ok {
		if s, ok := value.(string); ok {
			return s, nil
		}
		return "", fmt.Errorf("location: %s must be a string, got: %s", key, value)
	}
	return defaultValue, nil
}

// UnmarshalJSON decode a single object, stores its "type" field in the Type
// field, and the rest into the Args field
func (fl *FileLocation) UnmarshalJSON(data []byte) error {
	if err := json.Unmarshal(data, &fl.Args); err != nil {
		return err
	}
	if v, ok := fl.Args["type"]; ok {
		if s, ok := v.(string); ok {
			fl.Type = s
			delete(fl.Args, "type")
		} else {
			return fmt.Errorf("invalid type for the 'type' field. Expected a string, got: %s", v)
		}
	}
	return nil
}

// MarshalJSON encodes to json
func (fl FileLocation) MarshalJSON() ([]byte, error) {
	m := fl.Args
	if m == nil {
		m = map[string]interface{}{}
	}
	if fl.Type != "" {
		m["type"] = fl.Type
	}
	data, err := json.Marshal(m)
	delete(m, "type")
	return data, err
}

// FileMessageReader reads a file in a xbus message
type FileMessageReader struct {
	err  error
	meta *FileMeta
	c    <-chan api.Item
}

// NewFileMessageReader ...
func NewFileMessageReader(c <-chan api.Item) *FileMessageReader {
	return &FileMessageReader{
		c: c,
	}
}

// Metadata returns the file metadata
func (fm *FileMessageReader) Metadata() (*FileMeta, error) {
	if fm.err != nil {
		return nil, fm.err
	}
	if fm.meta != nil {
		return fm.meta, nil
	}
	head := <-fm.c
	var meta FileMeta
	if err := json.Unmarshal(head, &meta); err != nil {
		fm.err = err
		return nil, err
	}
	fm.meta = &meta
	return fm.meta, nil
}

// Body returns a io.ReadCloser on the file body. The reader must be closed
// to release the allocated resources
func (fm *FileMessageReader) Body() (io.ReadCloser, error) {
	metadata, err := fm.Metadata()
	if err != nil {
		return nil, err
	}
	if metadata.Location.Type == "embedded" || metadata.Location.Type == "" {
		return &BodyReader{c: fm.c}, nil
	}
	locationType := GetLocationType(metadata.Location.Type)
	if locationType == nil {
		return nil, fmt.Errorf("No such location type: %s", metadata.Location.Type)
	}
	return locationType.ReadBody(metadata.Location)
}

// BodyReader reads a file body embedded in a message
type BodyReader struct {
	c    <-chan api.Item
	next []byte
}

// Read implements io.Reader.Read
func (r *BodyReader) Read(buf []byte) (int, error) {
	if r.next == nil {
		if next, ok := <-r.c; ok {
			r.next = next
		} else {
			return 0, io.EOF
		}
	}
	n := copy(buf, r.next)
	if n < len(r.next) {
		r.next = r.next[n:]
	} else {
		r.next = nil
	}
	return n, nil
}

// Close does nothing
func (r *BodyReader) Close() error {
	return nil
}

// FileLocationType is implemented for each possible "location.type"
type FileLocationType interface {
	ReadBody(location FileLocation) (io.ReadCloser, error)
}

var fileLocationTypeRegistry = make(map[string]FileLocationType)

// RegisterFileLocationType registers a new FileLocationType
func RegisterFileLocationType(typename string, locationType FileLocationType) error {
	if _, ok := fileLocationTypeRegistry[typename]; ok {
		return fmt.Errorf("stdlib.RegisterFileLocationType: duplicate typename '%s'", typename)
	}
	fileLocationTypeRegistry[typename] = locationType
	return nil
}

// GetLocationType returns a FileLocationType
func GetLocationType(typename string) FileLocationType {
	t, _ := fileLocationTypeRegistry[typename]
	return t
}
