package routing

import (
	"sync"

	"orus.io/xbus/xbus-stdlib/service"
)

var (
	memoryTableStores      = make(map[string]*MemoryTableStore)
	memoryTableStoresMutex sync.Mutex
)

// MemoryTableStore stores key/values in a map
type MemoryTableStore struct {
	m     map[string]string
	mutex sync.Mutex
}

// NewMemoryTableStore returns a in-memory TableStore
func NewMemoryTableStore(storename string) *MemoryTableStore {
	memoryTableStoresMutex.Lock()
	defer memoryTableStoresMutex.Unlock()
	if kv, ok := memoryTableStores[storename]; ok {
		return kv
	}
	kv := MemoryTableStore{m: make(map[string]string)}
	memoryTableStores[storename] = &kv
	return &kv
}

// Set stores a value
func (s *MemoryTableStore) Set(key, value string) error {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	s.m[key] = value
	return nil
}

// Get stores a value
func (s *MemoryTableStore) Get(key string) (string, error) {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	if value, ok := s.m[key]; ok {
		return value, nil
	}
	return "", ErrNoSuchKey
}

// Delete the key from the table
func (s *MemoryTableStore) Delete(key string) error {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	delete(s.m, key)
	return nil
}

// Close does nothing
func (s *MemoryTableStore) Close() error {
	return nil
}

func memoryTableStoreFactory(prefix string, settings service.Settings) (TableStore, error) {
	storename, err := settings.String(prefix + "name")
	if err != nil {
		return nil, err
	}
	return NewMemoryTableStore(storename), nil
}

func init() {
	MustRegisterTableStore("memory", memoryTableStoreFactory)
}
