package service

import (
	"context"
	"fmt"

	"xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
)

// ConsumerFunc is a function implementing a consumer
type ConsumerFunc func(context.Context, api.ActorProcessRequest, Inputs, xbus.ZeroLogger) error

// Consumer ...
type Consumer interface {
	Process(context.Context, api.ActorProcessRequest, Inputs, xbus.ZeroLogger) error
}

// ConsumerFactory is a factory of Consumer
type ConsumerFactory func(*api.Actor, Settings) (Consumer, error)

// RegisterConsumer registers a consumer
func RegisterConsumer(name string, consumerFactory ConsumerFactory) (err error) {
	defer func() {
		if r := recover(); r != nil {
			err = fmt.Errorf("%v", r)
		}
	}()
	xbus.RegisterActorService(name, newConsumerServiceFactory(consumerFactory))
	return
}

// RegisterConsumerFunc registers a consumer
func RegisterConsumerFunc(name string, process ConsumerFunc) error {
	return RegisterConsumer(
		name,
		func(*api.Actor, Settings) (Consumer, error) {
			return process, nil
		})
}

// UnregisterConsumer unregisters a consumer
func UnregisterConsumer(name string) {
	xbus.UnregisterActorService(name)
}

// Process calls the function
func (fn ConsumerFunc) Process(ctx context.Context, req api.ActorProcessRequest, inputs Inputs, log xbus.ZeroLogger) error {
	return fn(ctx, req, inputs, log)
}

func newConsumerServiceFactory(factory ConsumerFactory) func(actor *xbus.Actor) xbus.ActorService {
	return func(actor *xbus.Actor) xbus.ActorService {
		apiActor := actor.Options.Actor
		settings := Settings(actor.Options.Service.Settings)
		consumer, err := factory(&apiActor, settings)
		if err != nil {
			panic(err)
		}
		return &actorService{
			actor:   actor,
			service: consumer,
			process: consumerProcess(consumer),
		}
	}
}

func consumerProcess(consumer Consumer) func(
	ctx context.Context, svc *actorService, req api.ActorProcessRequest, send func(api.ActorProcessingState),
) error {
	return func(ctx context.Context, svc *actorService, req api.ActorProcessRequest, send func(api.ActorProcessingState)) error {
		log := svc.actor.Client.Log.(xbus.ZeroLogger)
		inputs := realInputs{
			ctx:       ctx,
			client:    svc.actor.Client,
			actorID:   svc.actor.ID,
			actorKind: "consumer",
			inputs:    req.Inputs,
		}
		return consumer.Process(ctx, req, inputs, log)
	}
}
