package service

import (
	"errors"
	"fmt"
	"strings"
)

var (
	// ErrNoSuchKey is a err kind returned if the key is not present in the settings
	ErrNoSuchKey = fmt.Errorf("no such key")

	// ErrInvalidType is a err kind returnd if the setting value does not have the expected type
	ErrInvalidType = fmt.Errorf("invalid type")
)

// NoSuchKeyError is returned if the key is not present in the settings
type NoSuchKeyError struct {
	Key string
}

func (e NoSuchKeyError) Error() string {
	return fmt.Sprintf("no such key: %s", e.Key)
}

// Unwrap returns ErrNoSuchKey
func (e NoSuchKeyError) Unwrap() error {
	return ErrNoSuchKey
}

// InvalidTypeError is returned if the value has the wrong type
type InvalidTypeError struct {
	Key          string
	ExpectedType string
	Value        interface{}
}

func (e InvalidTypeError) Error() string {
	return fmt.Sprintf("invalid type for key '%s'. Expected a %s, got a %T: %v",
		e.Key, e.ExpectedType, e.Value, e.Value)
}

func (e InvalidTypeError) Unwrap() error {
	return ErrInvalidType
}

// Settings holds a Actor service settings
type Settings map[string]interface{}

// Get returns a setting value
func (s Settings) Get(key string) (interface{}, bool) {
	value, ok := s[key]
	return value, ok
}

// String returns a string setting value
func (s Settings) String(key string) (string, error) {
	value, ok := s.Get(key)
	if !ok {
		return "", NoSuchKeyError{key}
	}
	if s, ok := value.(string); ok {
		return s, nil
	}
	return "", InvalidTypeError{key, "string", value}
}

// StringD returns a string setting value, with a default value if the setting is missing
func (s Settings) StringD(key string, defaultValue string) (string, error) {
	value, err := s.String(key)
	if errors.Is(err, ErrNoSuchKey) {
		return defaultValue, nil
	}
	return value, err
}

// Bool returns a boolean setting
func (s Settings) Bool(key string) (bool, error) {
	value, ok := s.Get(key)
	if !ok {
		return false, NoSuchKeyError{key}
	}
	switch v := value.(type) {
	case bool:
		return v, nil
	case string:
		if v == "1" || strings.EqualFold(v, "true") || strings.EqualFold(v, "t") {
			return true, nil
		}
		if v == "0" || strings.EqualFold(v, "false") || strings.EqualFold(v, "f") {
			return false, nil
		}
		return false, InvalidTypeError{key, "bool", value}
	default:
		return false, InvalidTypeError{key, "bool", value}
	}
}

// BoolD returns a string setting value, with a default value if the setting is missing
func (s Settings) BoolD(key string, defaultValue bool) (bool, error) {
	value, err := s.Bool(key)
	if errors.Is(err, ErrNoSuchKey) {
		return defaultValue, nil
	}
	return value, err
}
