package xbus2xbus

import (
	"context"
	"fmt"
	"sync"
	"time"

	nrpc "github.com/nats-rpc/gogo-nrpc"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
)

// Consumer emits incoming envelope in another bus
type Consumer struct {
	log              xbus.Logger
	clientConfigFile string
	emitterName      string
	copyEnvelopeID   bool

	client    *xbus.Client
	emitterID api.UUID
}

// NewConsumer creates a Consumer
func NewConsumer(actor *xbus.Actor) xbus.ActorService {
	return &Consumer{
		log:              actor.GetClient().Log,
		clientConfigFile: actor.GetStringSetting("target-client-config"),
		emitterName:      actor.GetStringSetting("target-emitter"),
		copyEnvelopeID:   actor.GetBoolSettingD("copy-envelope-id", true),
	}
}

// Startup connects to the downside xbus
func (c *Consumer) Startup() error {
	options, err := xbus.COptionsFromFile(c.clientConfigFile, api.Account_ACTOR, c.log)
	if err != nil {
		return err
	}
	options.ForceStart = true
	client := xbus.NewClient(options, c.log)

	if err := client.Startup(false); err != nil {
		return err
	}
	c.client = client
	if err := c.loadEmitter(); err != nil {
		c.client = nil
		return err
	}
	return nil
}

// Shutdown disconnects from the downside xbus
func (c *Consumer) Shutdown() error {
	c.client.Shutdown()
	return nil
}

func (c *Consumer) loadEmitter() error {
	if c.client.GetConn() != nil && c.client.GetConn().IsConnected() {
		if emitter := c.client.GetActorByName(c.emitterName); emitter != nil {
			c.emitterID = emitter.ID
		} else {
			c.client.Shutdown()
			return fmt.Errorf("No such emitter: %s", c.emitterName)
		}
	}
	return nil
}

type eventItem struct {
	eventType string
	item      api.Item
}

// Process forwards the incoming envelope to the downside xbus, or returns
// a SERVERTOOBUSY error
func (c *Consumer) Process(apc *xbus.ActorProcessingContext) error {
	ctx, cancel := context.WithCancel(apc.Context)
	defer cancel()
	if !c.client.GetConn().IsConnected() {
		return &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[xbus2xbus] Not connected to downside xbus."),
		}
	}
	if c.emitterID.IsZero() {
		if err := c.loadEmitter(); err != nil {
			c.client = nil
			return err
		}
	}
	if c.emitterID.IsZero() {
		return &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[xbus2xbus] Cannot get downside emitter ID."),
		}
	}
	reader, err := apc.ReadEnvelope("default")
	if err != nil {
		return err
	}
	var eventIDs []api.UUID
	for _, event := range reader.Events() {
		eventIDs = append(eventIDs, event.ID)
	}

	envID := reader.ID
	if !c.copyEnvelopeID {
		envID, err = api.UUIDV4()
	}
	if err != nil {
		return err
	}
	output := xbus.NewLegacyEnvelopeSenderWithIDs(
		envID, reader.EventTypes(), eventIDs, api.EnvelopeMaxSize,
		func(env *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			ack, err := c.client.SendEnvelopeFragment(c.emitterID, nil, "default", *env)
			return ack, err
		},
	)

	var wg sync.WaitGroup
	items := make(chan eventItem)
	for _, eventType := range reader.EventTypes() {
		c := reader.GetItemChan(eventType)
		wg.Add(1)
		go func(eventType string, c <-chan api.Item) {
			defer wg.Done()
			for item := range c {
				select {
				case items <- eventItem{eventType, item}:
				case <-ctx.Done():
				}
			}
		}(eventType, c)
	}
	go func() {
		wg.Wait()
		close(items)
	}()
	timeout := time.AfterFunc(time.Minute, func() {
		cancel()
	})
	for item := range items {
		timeout.Reset(time.Minute)
		if err := output.AddItems(item.eventType, item.item); err != nil {
			return err
		}
	}
	if ctx.Err() != nil {
		reader.Cancel()
	}
	if ctx.Err() != nil || reader.Status() != api.EnvelopeAck_ACCEPTED {
		// TODO cancel emission
		//output.
		return fmt.Errorf("Envelope forward to downstream failed")
	}
	return output.Close()
}

func init() {
	xbus.RegisterActorService("xbus2xbus-consumer", NewConsumer)
}
